import unittest
from datetime import timedelta

from alldatetime.alldatetime import (
    _is_leap,
    _ord2ymd,
    _ymd2ord,
    alldate,
    alldateperiod,
    alldatetime,
    alltime,
)

connection = ""


class TestAllDateTime(unittest.TestCase):
    def test_all_date(self):
        timestamps = [
            ((1970, 1, 1), 0),
            ((1, 1, 1), -62135596800),
            ((-1, 12, 31), -62135683200),
            ((-1, 2, 29), -62162121600),
            ((-5, 2, 29), -62288352000),
            ((-4001, 2, 29), -188389929600),
        ]
        for date, timestamp in timestamps:
            ad = alldate(*date)
            self.assertEqual(int(timestamp), int(ad.timestamp))

            ad = alldate.fromtimestamp(timestamp)
            self.assertEqual(date, (ad.year, ad.month, ad.day))
            self.assertEqual(int(timestamp), int(ad.timestamp))

    def test_all_datetime(self):
        timestamps = [
            ((1970, 1, 1, 0, 0, 0, 0), 0),
            ((1970, 1, 1, 0, 0, 1, 0), 1),
            ((-1234, 8, 1, 14, 29, 18, 0), -101058571842),
            ((-4001, 1, 29, 15, 51, 30, 0), -188392550910),
            ((-5, 2, 29, 15, 51, 30, 0), -62288294910),
            ((-1, 12, 31, 23, 59, 59, 0), -62135596801),
            ((1, 1, 1, 0, 0, 0, 0), -62135596800),
            ((2023, 12, 29, 15, 51, 30, 0), 1703865090),
        ]
        for datetime, timestamp in timestamps:
            adt = alldatetime(*datetime)
            self.assertEqual(int(adt.timestamp), int(timestamp))

            adt = alldatetime.fromtimestamp(timestamp)
            self.assertEqual(
                datetime,
                (
                    adt.year,
                    adt.month,
                    adt.day,
                    adt.hour,
                    adt.minute,
                    adt.second,
                    adt.microsecond,
                ),
            )
            self.assertEqual(int(adt.timestamp), int(timestamp))

    def test_alldate_comparisons(self):
        dates = [
            # alldate1, alldate2, equal, less than, less than or equal
            (alldate(1, 1, 1), alldate(1, 1, 1), True, False, True),
            (alldate(1, 1, 2), alldate(1, 1, 1), False, False, False),
            (alldate(1, 2, 1), alldate(1, 1, 1), False, False, False),
            (alldate(2, 1, 1), alldate(1, 1, 1), False, False, False),
            (alldate(1, 1, 1), alldate(1, 1, 2), False, True, True),
            (alldate(1, 1, 1), alldate(1, 2, 1), False, True, True),
            (alldate(1, 1, 1), alldate(2, 1, 1), False, True, True),
        ]
        for alldate1, alldate2, equal, lt, le in dates:
            self.assertEqual(alldate1 == alldate2, equal)
            self.assertEqual(alldate1 < alldate2, lt)
            self.assertEqual(alldate1 <= alldate2, le)
            self.assertEqual(alldate1 > alldate2, not le)
            self.assertEqual(alldate1 >= alldate2, not lt)

            self.assertEqual(alldate2 == alldate1, equal)
            self.assertEqual(alldate2 < alldate1, not le)
            self.assertEqual(alldate2 <= alldate1, not lt)
            self.assertEqual(alldate2 > alldate1, lt)
            self.assertEqual(alldate2 >= alldate1, le)

    def test_alldate_add_timedelta(self):
        add_dates = [
            (alldate(-1201, 2, 1), timedelta(days=28), alldate(-1201, 2, 29)),
            (alldate(-1, 12, 1), timedelta(days=1), alldate(-1, 12, 2)),
            (alldate(-1, 12, 1), timedelta(days=30), alldate(-1, 12, 31)),
            (alldate(-1, 12, 1), timedelta(days=31), alldate(1, 1, 1)),
            (alldate(1999, 2, 1), timedelta(days=28), alldate(1999, 3, 1)),
            (alldate(1999, 2, 1), timedelta(days=1), alldate(1999, 2, 2)),
            (alldate(1999, 2, 1), timedelta(days=28), alldate(1999, 3, 1)),
            (alldate(2000, 2, 1), timedelta(days=1), alldate(2000, 2, 2)),
            (alldate(2000, 2, 1), timedelta(days=28), alldate(2000, 2, 29)),
            (alldate(2023, 12, 1), timedelta(days=1), alldate(2023, 12, 2)),
            (alldate(2023, 12, 1), timedelta(days=30), alldate(2023, 12, 31)),
            (alldate(2023, 12, 1), timedelta(days=31), alldate(2024, 1, 1)),
        ]

        for date, delta, result in add_dates:
            self.assertEqual(date + delta, result)
            self.assertEqual(result - delta, date)

    def test_alltime_comparisons(self):
        dates = [
            # alltime1, alltime2, equal, less than, less than or equal
            (alltime(14, 10, 10, 1), alltime(14, 10, 10, 1), True, False, True),
            (alltime(14, 10, 10, 2), alltime(14, 10, 10, 1), False, False, False),
            (alltime(14, 10, 11, 1), alltime(14, 10, 10, 1), False, False, False),
            (alltime(14, 11, 10, 1), alltime(14, 10, 10, 1), False, False, False),
            (alltime(15, 10, 10, 1), alltime(14, 10, 10, 1), False, False, False),
            (alltime(14, 10, 10, 1), alltime(14, 10, 10, 2), False, True, True),
            (alltime(14, 10, 10, 1), alltime(14, 10, 11, 1), False, True, True),
            (alltime(14, 10, 10, 1), alltime(14, 11, 10, 1), False, True, True),
            (alltime(14, 10, 10, 1), alltime(15, 10, 10, 1), False, True, True),
        ]
        for alltime1, alltime2, equal, lt, le in dates:
            self.assertEqual(alltime1 == alltime2, equal)
            self.assertEqual(alltime1 < alltime2, lt)
            self.assertEqual(alltime1 <= alltime2, le)
            self.assertEqual(alltime1 > alltime2, not le)
            self.assertEqual(alltime1 >= alltime2, not lt)

            self.assertEqual(alltime2 == alltime1, equal)
            self.assertEqual(alltime2 < alltime1, not le)
            self.assertEqual(alltime2 <= alltime1, not lt)
            self.assertEqual(alltime2 > alltime1, lt)
            self.assertEqual(alltime2 >= alltime1, le)

    def test_alldatetime_comparisons(self):
        dates = [
            # alldatetime1, alldatetime2, equal, less than, less than or equal
            (
                alldatetime(1, 1, 1, 14, 10, 10, 1),
                alldatetime(1, 1, 1, 14, 10, 10, 1),
                True,
                False,
                True,
            ),
            (
                alldatetime(1, 1, 1, 14, 10, 10, 2),
                alldatetime(1, 1, 1, 14, 10, 10, 1),
                False,
                False,
                False,
            ),
            (
                alldatetime(1, 1, 2, 14, 10, 10, 1),
                alldatetime(1, 1, 1, 14, 10, 10, 1),
                False,
                False,
                False,
            ),
            (
                alldatetime(1, 1, 1, 14, 10, 10, 1),
                alldatetime(1, 1, 1, 14, 10, 10, 2),
                False,
                True,
                True,
            ),
            (
                alldatetime(1, 1, 1, 14, 10, 10, 1),
                alldatetime(1, 1, 2, 14, 10, 10, 1),
                False,
                True,
                True,
            ),
        ]
        for alldatetime1, alldatetime2, equal, lt, le in dates:
            self.assertEqual(alldatetime1 == alldatetime2, equal)
            self.assertEqual(alldatetime1 < alldatetime2, lt)
            self.assertEqual(alldatetime1 <= alldatetime2, le)
            self.assertEqual(alldatetime1 > alldatetime2, not le)
            self.assertEqual(alldatetime1 >= alldatetime2, not lt)

            self.assertEqual(alldatetime2 == alldatetime1, equal)
            self.assertEqual(alldatetime2 < alldatetime1, not le)
            self.assertEqual(alldatetime2 <= alldatetime1, not lt)
            self.assertEqual(alldatetime2 > alldatetime1, lt)
            self.assertEqual(alldatetime2 >= alldatetime1, le)

    def test_strftime(self):
        date_formats = [
            (alldate(-5000, 1, 8), "5000-01-08 BC", "5000/01/08 BC"),
            (alldate(-5000, 11, 8), "5000-11-08 BC", "5000/11/08 BC"),
            (alldate(2000, 1, 8), "2000-01-08 AD", "2000/01/08 AD"),
        ]
        for d, s1, s2 in date_formats:
            self.assertEqual(d.strftime("%Y-%m-%d"), s1)
            self.assertEqual(d.strftime("%Y/%m/%d"), s2)

        time_formats = [
            (alltime(1, 15, 30, 1541), "01:15:30.001541", "01:15:30"),
            (alltime(23, 5, 30, 0), "23:05:30.000000", "23:05:30"),
        ]
        for t, s1, s2 in time_formats:
            self.assertEqual(t.strftime("%H:%M:%S.%f"), s1)
            self.assertEqual(t.strftime("%H:%M:%S"), s2)

        datetime_formats = [
            (
                alldatetime(-5000, 1, 8, 8, 30, 15),
                "5000-01-08 08:30:15 BC",
                "5000/01/08 08:30:15 BC",
            ),
            (
                alldatetime(2000, 1, 8, 8, 30, 15),
                "2000-01-08 08:30:15 AD",
                "2000/01/08 08:30:15 AD",
            ),
        ]
        for dt, s1, s2 in datetime_formats:
            self.assertEqual(dt.strftime("%Y-%m-%d %H:%M:%S"), s1)
            self.assertEqual(dt.strftime("%Y/%m/%d %H:%M:%S"), s2)

    def test_strptime(self):
        datetime_formats = [
            (
                "5000-01-08 08:30:15 BC",
                "5000/01/08 08:30:15 BC",
                alldatetime(-5000, 1, 8, 8, 30, 15),
            ),
            (
                "2000-01-08 08:30:15 AD",
                "2000/01/08 08:30:15 AD",
                alldatetime(2000, 1, 8, 8, 30, 15),
            ),
            (
                "2000-01-08 08:30:15",
                "2000/01/08 08:30:15",
                alldatetime(2000, 1, 8, 8, 30, 15),
            ),
        ]
        for s1, s2, dt in datetime_formats:
            self.assertEqual(alldatetime.strptime(s1, "%Y-%m-%d %H:%M:%S"), dt)
            self.assertEqual(alldatetime.strptime(s2, "%Y/%m/%d %H:%M:%S"), dt)

        datetime_formats = [
            ("5000 BC", alldatetime(-5000, 1, 1, 0, 0, 0)),
            ("2000 AD", alldatetime(2000, 1, 1, 0, 0, 0)),
        ]
        for s, dt in datetime_formats:
            self.assertEqual(alldatetime.strptime(s, "%Y"), dt)

    def test__ymd2ord(self):
        ymd2ords = [
            ((1, 1, 1), 0),
            ((1, 1, 2), 1),
            ((1, 12, 31), 364),
            ((2, 1, 1), 365),
            ((1600, 12, 31), 584387),
            ((1601, 1, 1), 584388),
            ((2023, 12, 14), 738867),
            ((-1, 12, 31), -1),
            ((-1, 12, 30), -2),
            ((-1, 1, 1), -366),
            ((-2, 12, 31), -367),
            ((-5, 12, 31), -1462),
            ((-6, 12, 31), -1828),
        ]

        for ymd, ord in ymd2ords:
            year, month, day = ymd
            self.assertEqual(_ymd2ord(year, month, day), ord)
            self.assertEqual(_ord2ymd(ord), ymd)

    def test__is_leap(self):
        leaps = [
            (1, False),
            (4, True),
            (100, False),
            (400, True),
            (1956, True),
            (2008, True),
            (-1, True),
            (-2, False),
            (-5, True),
            (-100, False),
            (-101, False),
            (-201, False),
            (-300, False),
            (-301, False),
            (-401, True),
        ]

        for year, leap in leaps:
            self.assertEqual(_is_leap(year), leap)

    def test_alldateperiod(self):
        periods = [
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 10, 10), alldate(2023, 11, 30)),
                False,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 10, 10), alldate(2023, 12, 1)),
                False,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 10, 10), alldate(2023, 12, 3)),
                True,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 3)),
                True,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 10, 10), alldate(2023, 12, 31)),
                True,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 31)),
                True,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 12, 5), alldate(2023, 12, 31)),
                False,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldateperiod(alldate(2023, 12, 6), alldate(2023, 12, 31)),
                False,
            ),
        ]

        for period1, period2, overlap in periods:
            self.assertEqual(period1.overlap_with(period2), overlap)
            self.assertEqual(period2.overlap_with(period1), overlap)

        with self.assertRaises(ValueError):
            alldateperiod(None, alldate(202, 1, 1))
        with self.assertRaises(ValueError):
            alldateperiod(alldate(202, 1, 2), alldate(202, 1, 1))

        periods = [
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldate(2023, 10, 30),
                False,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldate(2023, 12, 1),
                True,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldate(2023, 12, 4),
                True,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldate(2023, 12, 5),
                False,
            ),
            (
                alldateperiod(alldate(2023, 12, 1), alldate(2023, 12, 5)),
                alldate(2023, 12, 31),
                False,
            ),
        ]
        for period, date, cover in periods:
            self.assertEqual(period.cover(date), cover)



from datetime import timedelta
from enum import Enum
from operator import index as _index

from alldatetime.alldatetime import (
    _check_month,
    _check_year,
    _days_in_month,
    alldate,
    alldateperiod,
)


class PrecisionUnit(Enum):
    Year = "Year"
    Month = "Month"
    Day = "Day"
    # Hour = "Hour"
    # Minute = "Minute"
    # Second = "Second"


class Precision:
    def __init__(self, num: int, unit: PrecisionUnit):
        if num < 0:
            raise ValueError("num should not be negative.")
        self._num = num
        self._unit = unit

    @property
    def num(self):
        return self._num

    @property
    def unit(self):
        return self._unit


class fuzzydate:
    def __init__(
        self,
        year: int = None,
        month: int = None,
        day: int = None,
        precision: Precision = None,
        forward_precision: Precision = None,
        backward_precision: Precision = None,
    ):
        (
            self._year,
            self._month,
            self._day,
            self._precision,
            self._forward_precision,
            self._backward_precision,
        ) = self._check_parameters(
            year, month, day, precision, forward_precision, backward_precision
        )
        self._anchor = alldate(self._year, self._month or 1, self._day or 1)

    def _check_parameters(
        self,
        year: int = None,
        month: int = None,
        day: int = None,
        precision: Precision = None,
        forward_precision: Precision = None,
        backward_precision: Precision = None,
    ):
        if year is None:
            raise ValueError("year should not be None.")
        _check_year(year)
        inferred_forward_precision = Precision(0, PrecisionUnit.Year)
        inferred_backward_precision = Precision(1, PrecisionUnit.Year)
        if month is not None:
            _check_month(month)
            inferred_forward_precision = Precision(0, PrecisionUnit.Month)
            inferred_backward_precision = Precision(1, PrecisionUnit.Month)
        if day is not None:
            if month is None:
                raise ValueError("month should not be None if day is not None.")
            day = _index(day)
            dim = _days_in_month(year, month)
            if not 1 <= day <= dim:
                raise ValueError(f"day must be in 1..{dim}")
            inferred_forward_precision = Precision(0, PrecisionUnit.Day)
            inferred_backward_precision = Precision(1, PrecisionUnit.Day)
        if forward_precision is None:
            forward_precision = precision or inferred_forward_precision
        if backward_precision is None:
            backward_precision = precision or inferred_backward_precision

        return (
            year,
            month,
            day,
            precision,
            forward_precision,
            backward_precision,
        )

    @property
    def year(self):
        return self._year

    @property
    def month(self):
        return self._month

    @property
    def day(self):
        return self._day

    @property
    def anchor(self):
        return self._anchor

    @property
    def forward_precision(self):
        return self._forward_precision

    @property
    def backward_precision(self):
        return self._backward_precision

    def to_alldateperiod(self) -> alldateperiod:
        forward_timedelta = self._precision_to_timedelta(self.forward_precision)
        backward_timedelta = self._precision_to_timedelta(self.backward_precision)
        return alldateperiod(
            self.anchor - forward_timedelta, self.anchor + backward_timedelta
        )

    def _precision_to_timedelta(self, precision: Precision):
        if precision.unit == PrecisionUnit.Year:
            return timedelta(days=precision.num * 365)
        if precision.unit == PrecisionUnit.Month:
            return timedelta(days=int(precision.num * 30.5))
        if precision.unit == PrecisionUnit.Day:
            return timedelta(days=precision.num)

    def to_alldateperiod_timestamps(self) -> tuple[float, float]:
        period = self.to_alldateperiod()
        return (period.start_date.timestamp, period.end_date.timestamp)

    def overlap_with(self, other) -> bool:
        if not isinstance(other, fuzzydate):
            return False
        self_period = self.to_alldateperiod()
        other_period = other.to_alldateperiod()
        return self_period.overlap_with(other_period)


class fuzzydateperiod:
    __slots__ = "_start_date", "_end_date", "_hashcode"

    def __init__(self, start_date: fuzzydate, end_date: fuzzydate):
        if start_date is None:
            raise ValueError("start_date should not be None.")
        if end_date is None:
            raise ValueError("end_date should not be None.")
        if start_date.anchor > end_date.anchor:
            raise ValueError("start_date should be earlier than end_date.")
        self._start_date = start_date
        self._end_date = end_date
        self._hashcode = -1

    @property
    def start_date(self):
        return self._start_date

    @property
    def end_date(self):
        return self._end_date

    def cover(self, date: fuzzydate):
        if date is None:
            raise ValueError("date should not be None.")
        start_period = self._start_date.to_alldateperiod()
        end_period = self._end_date.to_alldateperiod()
        date_period = date.to_alldateperiod()
        start_date = start_period.start_date
        end_date = end_period.end_date
        return not (
            start_date >= date_period.end_date or end_date <= date_period.start_date
        )



import math as _math
import time as _time
from datetime import date, datetime, time, timedelta
from operator import index as _index

__all__ = ("alldate", "alltime", "alldatetime")


def _cmp(x, y):
    return 0 if x == y else 1 if x > y else -1


BCENDING = " BC"
ADENDING = " AD"

_MONTHNAMES = [
    None,
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
]
_DAYNAMES = [None, "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
# -1 is a placeholder for indexing purposes.
_DAYS_IN_MONTH = [-1, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
_DAYS_BEFORE_MONTH = [-1]  # -1 is a placeholder for indexing purposes.

SECONDSPERHOUR = 3600
SECONDSPERMINUTE = 60

dbm = 0
for dim in _DAYS_IN_MONTH[1:]:
    _DAYS_BEFORE_MONTH.append(dbm)
    dbm += dim
del dbm, dim


def _check_year(year: int) -> int:
    year = _index(year)
    if year == 0:
        raise ValueError("year should not be 0", year)
    return year


def _check_month(month: int) -> int:
    month = _index(month)
    if not 1 <= month <= 12:
        raise ValueError("month must be in 1..12", month)
    return month


def _is_leap(year: int):
    "year -> 1 if leap year, else 0."
    if year < 0:
        year += 1
        year = -year
    return year % 4 == 0 and (year % 100 != 0 or year % 400 == 0)


def _days_before_year(year):
    "year -> number of days before January 1st of year."
    if year > 0:
        y = year - 1
        return y * 365 + y // 4 - y // 100 + y // 400
    if year < 0:
        y = -year
        return -(y * 365 + 1 + (y - 1) // 4 - (y - 1) // 100 + (y - 1) // 400)


def _days_in_month(year, month):
    "year, month -> number of days in that month in that year."
    month = _check_month(month)
    if month == 2 and _is_leap(year):
        return 29
    return _DAYS_IN_MONTH[month]


def _days_before_month(year, month):
    "year, month -> number of days in year preceding first day of month."
    month = _check_month(month)
    return _DAYS_BEFORE_MONTH[month] + (month > 2 and _is_leap(year))


_DI400Y = _days_before_year(401)  # number of days in 400 years
_DI100Y = _days_before_year(101)  #    "    "   "   " 100   "
_DI4Y = _days_before_year(5)  #    "    "   "   "   4   "


def _ymd2ord(year: int, month: int, day: int) -> int:
    "year, month, day -> ordinal, considering 01-Jan-0001 as day 0."
    year, month, day = _check_date_fields(year, month, day)
    return _days_before_year(year) + _days_before_month(year, month) + day - 1


def _ord2ymd(n):
    "ordinal -> (year, month, day), considering 01-Jan-0001 as day 0."

    # n is a 0-based index, starting at 1-Jan-1.  The pattern of leap years
    # repeats exactly every 400 years.  The basic strategy is to find the
    # closest 400-year boundary at or before n, then work with the offset
    # from that boundary to n.  Life is much clearer if we subtract 1 from
    # n first -- then the values of n at 400-year boundaries are exactly
    # those divisible by _DI400Y:
    #
    #     D  M   Y            n              n-1
    #     -- --- ----        ----------     ----------------
    #     31 Dec -400        -_DI400Y       -_DI400Y -1
    #      1 Jan -399         -_DI400Y +1   -_DI400Y      400-year boundary
    #     ...
    #     30 Dec  000        -1             -2
    #     31 Dec  000         0             -1
    #      1 Jan  001         1              0            400-year boundary
    #      2 Jan  001         2              1
    #      3 Jan  001         3              2
    #     ...
    #     31 Dec  400         _DI400Y        _DI400Y -1
    #      1 Jan  401         _DI400Y +1     _DI400Y      400-year boundary
    # n -= 1
    n400, n = divmod(n, _DI400Y)
    if n400 >= 0:
        year = n400 * 400 + 1  # ..., -399, 1, 401, ...
    else:
        year = n400 * 400

    # Now n is the (non-negative) offset, in days, from January 1 of year, to
    # the desired date.  Now compute how many 100-year cycles precede n.
    # Note that it's possible for n100 to equal 4!  In that case 4 full
    # 100-year cycles precede the desired day, which implies the desired
    # day is December 31 at the end of a 400-year cycle.
    n100, n = divmod(n, _DI100Y)

    # Now compute how many 4-year cycles precede it.
    n4, n = divmod(n, _DI4Y)

    # And now how many single years.  Again n1 can be 4, and again meaning
    # that the desired day is December 31 at the end of the 4-year cycle.
    n1, n = divmod(n, 365)

    year += n100 * 100 + n4 * 4 + n1
    if n1 == 4 or n100 == 4:
        assert n == 0
        return year - 1, 12, 31

    # Now the year is correct, and n is the offset from January 1.  We find
    # the month via an estimate that's either exact or one too large.
    leapyear = n1 == 3 and (n4 != 24 or n100 == 3)
    assert leapyear == _is_leap(year)
    month = (n + 50) >> 5
    preceding = _DAYS_BEFORE_MONTH[month] + (month > 2 and leapyear)
    if preceding > n:  # estimate is too large
        month -= 1
        preceding -= _DAYS_IN_MONTH[month] + (month == 2 and leapyear)
    n -= preceding
    assert 0 <= n < _days_in_month(year, month)

    # Now the year and month are correct, and n is the offset from the
    # start of that month:  we're done!
    return year, month, n + 1


def _build_struct_time(y, m, d, hh, mm, ss, dstflag):
    wday = (_ymd2ord(y, m, d) + 7) % 7
    dnum = _days_before_month(y, m) + d
    return _time.struct_time((y, m, d, hh, mm, ss, wday, dnum, dstflag))


def _days_in_month(year, month):
    "year, month -> number of days in that month in that year."
    month = _check_month(month)
    if month == 2 and _is_leap(year):
        return 29
    return _DAYS_IN_MONTH[month]


def _check_date_fields(year: int, month: int, day: int) -> tuple[int, int, int]:
    year = _check_year(year)
    month = _check_month(month)
    day = _index(day)
    dim = _days_in_month(year, month)
    if not 1 <= day <= dim:
        raise ValueError("day must be in 1..%d" % dim, day)
    return year, month, day


def _check_time_fields(hour, minute, second, microsecond):
    hour = _index(hour)
    minute = _index(minute)
    second = _index(second)
    microsecond = _index(microsecond)
    if not 0 <= hour <= 23:
        raise ValueError("hour must be in 0..23", hour)
    if not 0 <= minute <= 59:
        raise ValueError("minute must be in 0..59", minute)
    if not 0 <= second <= 59:
        raise ValueError("second must be in 0..59", second)
    if not 0 <= microsecond <= 999999:
        raise ValueError("microsecond must be in 0..999999", microsecond)
    return hour, minute, second, microsecond


class alldate:
    __slots__ = "_year", "_month", "_day", "_hashcode", "_timestamp"

    def __init__(self, year: int, month: int, day: int):
        year, month, day = _check_date_fields(year, month, day)
        self._year = year
        self._month = month
        self._day = day
        if year < 0:
            year += 1
        self._timestamp = _time.mktime((year, month, day, 0, 0, 0, 0, 0, 0))
        self._hashcode = -1

    @classmethod
    def fromtimestamp(cls, timestamp: int):
        "Construct a date from a POSIX timestamp (like time.time())."
        y, m, d, hh, mm, ss, weekday, jday, dst = _time.gmtime(timestamp)
        if y <= 0:
            y -= 1
        return cls(y, m, d)

    @classmethod
    def today(cls):
        "Construct a date from time.time()."
        t = _time.time()
        return cls.fromtimestamp(t)

    @classmethod
    def fromordinal(cls, n: int):
        """Construct a date from a proleptic Gregorian ordinal.

        January 1 of year 1 is day 0.  Only the year, month and day are
        non-zero in the result.
        """
        y, m, d = _ord2ymd(n)
        return cls(y, m, d)

    @property
    def year(self) -> int:
        """year (1-9999)"""
        return self._year

    @property
    def month(self) -> int:
        """month (1-12)"""
        return self._month

    @property
    def day(self) -> int:
        """day (1-31)"""
        return self._day

    @property
    def timestamp(self) -> float:
        return self._timestamp

    # Comparisons of date objects with other.

    def __eq__(self, other):
        if isinstance(other, alldate):
            return self._cmp(other) == 0
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, alldate):
            return self._cmp(other) <= 0
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, alldate):
            return self._cmp(other) < 0
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, alldate):
            return self._cmp(other) >= 0
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, alldate):
            return self._cmp(other) > 0
        return NotImplemented

    def _cmp(self, other):
        assert isinstance(other, alldate)
        y, m, d = self._year, self._month, self._day
        y2, m2, d2 = other._year, other._month, other._day
        return _cmp((y, m, d), (y2, m2, d2))

    def __hash__(self):
        "Hash."
        if self._hashcode == -1:
            self._hashcode = hash(self._getstate())
        return self._hashcode

    def _getstate(self):
        yhi, ylo = divmod(self._year, 256)
        return (bytes([yhi, ylo, self._month, self._day]),)

    def toordinal(self):
        """Return proleptic Gregorian ordinal for the year, month and day.

        January 1 of year 1 is day 0.  Only the year, month and day values
        contribute to the result.
        """
        return _ymd2ord(self._year, self._month, self._day)

    def __add__(self, other):
        "Add a date to a timedelta."
        if isinstance(other, timedelta):
            o = self.toordinal() + other.days
            return type(self).fromordinal(o)
        return NotImplemented

    __radd__ = __add__

    def __sub__(self, other):
        """Subtract two dates, or a date and a timedelta."""
        if isinstance(other, timedelta):
            return self + timedelta(-other.days)
        if isinstance(other, alldate):
            days1 = self.toordinal()
            days2 = other.toordinal()
            return timedelta(days1 - days2)
        return NotImplemented

    def isoformat(self):
        """Return the date formatted according to ISO.

        This is 'YYYY-MM-DD'.

        References:
        - http://www.w3.org/TR/NOTE-datetime
        - http://www.cl.cam.ac.uk/~mgk25/iso-time.html
        """
        formater = "%04d-%02d-%02d" if self._year > 0 else "%05d-%02d-%02d"
        return formater % (self._year, self._month, self._day)

    __str__ = isoformat

    def strftime(self, format):
        date_string = date(abs(self.year), self.month, self.day).strftime(format)
        if self.year < 0:
            date_string += BCENDING
        else:
            date_string += ADENDING

        return date_string


class alldateperiod:
    """
    Represents a time interval, consisting of a start time and an end time, forming an open-closed interval.
    """

    __slots__ = "_start_date", "_end_date", "_hashcode"

    def __init__(self, start_date: alldate, end_date: alldate):
        if start_date is None:
            raise ValueError("start_date should not be None.")
        if end_date is None:
            raise ValueError("end_date should not be None.")
        if start_date > end_date:
            raise ValueError("start_date should be earlier than end_date.")
        self._start_date = start_date
        self._end_date = end_date
        self._hashcode = -1

    @property
    def start_date(self):
        return self._start_date

    @property
    def end_date(self):
        return self._end_date

    def overlap_with(self, other) -> bool:
        if not isinstance(other, alldateperiod):
            return False
        return not (
            self.end_date <= other.start_date or self.start_date >= other.end_date
        )
    
    def cover(self, date: alldate) -> bool:
        if not isinstance(date, alldate):
            raise ValueError("date should be of type alldate.")
        return date >= self._start_date and date < self._end_date

    def __eq__(self, other):
        if isinstance(other, alldateperiod):
            return (
                self.start_date == other.start_date and self.end_date == other.end_date
            )
        else:
            return NotImplemented

    def __hash__(self):
        if self._hashcode == -1:
            t = self
            self._hashcode = hash(t._getstate()[0])
        return self._hashcode

    def _getstate(self):
        yhi1, ylo1 = divmod(self.start_date.year, 256)
        yhi2, ylo2 = divmod(self.end_date.year, 256)
        return (
            bytes(
                [
                    yhi1,
                    ylo1,
                    self.start_date.month,
                    self.start_date.day,
                    yhi2,
                    ylo2,
                    self.end_date.month,
                    self.end_date.day,
                ]
            ),
        )


def _format_time(hh, mm, ss, us, timespec="auto"):
    specs = {
        "hours": "{:02d}",
        "minutes": "{:02d}:{:02d}",
        "seconds": "{:02d}:{:02d}:{:02d}",
        "milliseconds": "{:02d}:{:02d}:{:02d}.{:03d}",
        "microseconds": "{:02d}:{:02d}:{:02d}.{:06d}",
    }

    if timespec == "auto":
        # Skip trailing microseconds when us==0.
        timespec = "microseconds" if us else "seconds"
    elif timespec == "milliseconds":
        us //= 1000
    try:
        fmt = specs[timespec]
    except KeyError:
        raise ValueError("Unknown timespec value")
    else:
        return fmt.format(hh, mm, ss, us)


class alltime:
    __slots__ = "_hour", "_minute", "_second", "_microsecond", "_hashcode"

    def __init__(self, hour, minute, second, microsecond=0):
        self._hour, self._minute, self._second, self._microsecond = _check_time_fields(
            hour, minute, second, microsecond
        )
        self._hashcode = -1

    @property
    def hour(self):
        """hour (0-23)"""
        return self._hour

    @property
    def minute(self):
        """minute (0-59)"""
        return self._minute

    @property
    def second(self):
        """second (0-59)"""
        return self._second

    @property
    def microsecond(self):
        """microsecond (0-999999)"""
        return self._microsecond

    def seconds_from_zero_hour(self):
        return (
            self._hour * SECONDSPERHOUR
            + self._minute * SECONDSPERMINUTE
            + self._second
            + self._microsecond / 1e6
        )

    def __eq__(self, other):
        if isinstance(other, alltime):
            return self._cmp(other) == 0
        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, alltime):
            return self._cmp(other) <= 0
        else:
            return NotImplemented

    def __lt__(self, other):
        if isinstance(other, alltime):
            return self._cmp(other) < 0
        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, alltime):
            return self._cmp(other) >= 0
        else:
            return NotImplemented

    def __gt__(self, other):
        if isinstance(other, alltime):
            return self._cmp(other) > 0
        else:
            return NotImplemented

    def _cmp(self, other):
        assert isinstance(other, alltime)

        return _cmp(
            (self._hour, self._minute, self._second, self._microsecond),
            (other._hour, other._minute, other._second, other._microsecond),
        )

    def __hash__(self):
        """Hash."""
        if self._hashcode == -1:
            t = self
            self._hashcode = hash(t._getstate()[0])
        return self._hashcode

    def _getstate(self, protocol=3):
        us2, us3 = divmod(self._microsecond, 256)
        us1, us2 = divmod(us2, 256)
        h = self._hour
        basestate = bytes([h, self._minute, self._second, us1, us2, us3])
        return (basestate,)

    def isoformat(self, timespec="auto"):
        """Return the time formatted according to ISO.

        The full format is 'HH:MM:SS.mmmmmm+zz:zz'. By default, the fractional
        part is omitted if self.microsecond == 0.

        The optional argument timespec specifies the number of additional
        terms of the time to include. Valid options are 'auto', 'hours',
        'minutes', 'seconds', 'milliseconds' and 'microseconds'.
        """
        s = _format_time(
            self._hour, self._minute, self._second, self._microsecond, timespec
        )

        return s

    __str__ = isoformat

    def strftime(self, format):
        return time(self.hour, self.minute, self.second, self.microsecond).strftime(
            format
        )


class alldatetime:
    __slots__ = ("_date", "_time", "_hashcode")

    def __init__(
        self,
        year: int,
        month: int,
        day: int,
        hour: int = 0,
        minute: int = 0,
        second: int = 0,
        microsecond: int = 0,
    ):
        self._date = alldate(year, month, day)
        self._time = alltime(hour, minute, second, microsecond)
        self._hashcode = -1

    @classmethod
    def fromtimestamp(cls, timestamp: int):
        frac, timestamp = _math.modf(timestamp)
        us = round(frac * 1e6)
        if us >= 1000000:
            timestamp += 1
            us -= 1000000
        elif us < 0:
            timestamp -= 1
            us += 1000000
        y, m, d, hh, mm, ss, weekday, jday, dst = _time.gmtime(timestamp)
        if y <= 0:
            y -= 1
        return cls(y, m, d, hh, mm, ss, us)

    @property
    def year(self):
        """year (1-9999)"""
        return self._date.year

    @property
    def month(self):
        """month (1-12)"""
        return self._date.month

    @property
    def day(self):
        """day (1-31)"""
        return self._date.day

    @property
    def hour(self):
        """hour (0-23)"""
        return self._time.hour

    @property
    def minute(self):
        """minute (0-59)"""
        return self._time.minute

    @property
    def second(self):
        """second (0-59)"""
        return self._time.second

    @property
    def microsecond(self):
        """microsecond (0-999999)"""
        return self._time.microsecond

    @property
    def timestamp(self):
        return self._date.timestamp + self._time.seconds_from_zero_hour()

    def date(self) -> alldate:
        "Return the date part."
        return alldate(self._date.year, self._date.month, self._date.day)

    def time(self) -> alltime:
        return alltime(
            self._time.hour,
            self._time.minute,
            self._time.second,
            self._time.microsecond,
        )

    def __eq__(self, other):
        if isinstance(other, alldatetime):
            return self._cmp(other) == 0
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, alldatetime):
            return self._cmp(other) <= 0
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, alldatetime):
            return self._cmp(other) < 0
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, alldatetime):
            return self._cmp(other) >= 0
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, alldatetime):
            return self._cmp(other) > 0
        return NotImplemented

    def _cmp(self, other):
        assert isinstance(other, alldatetime)
        date_cmp = self._date._cmp(other._date)
        if date_cmp == 0:
            return self._time._cmp(other._time)
        return date_cmp

    def __hash__(self):
        if self._hashcode == -1:
            t = self
            self._hashcode = hash(t._getstate()[0])
        return self._hashcode

    def _getstate(self, protocol=3):
        yhi, ylo = divmod(self._year, 256)
        us2, us3 = divmod(self._microsecond, 256)
        us1, us2 = divmod(us2, 256)
        m = self._month
        basestate = bytes(
            [
                yhi,
                ylo,
                m,
                self._day,
                self._hour,
                self._minute,
                self._second,
                us1,
                us2,
                us3,
            ]
        )
        return (basestate,)

    def isoformat(self, timespec="auto"):
        """Return the time formatted according to ISO.

        The full format is 'HH:MM:SS.mmmmmm+zz:zz'. By default, the fractional
        part is omitted if self.microsecond == 0.

        The optional argument timespec specifies the number of additional
        terms of the time to include. Valid options are 'auto', 'hours',
        'minutes', 'seconds', 'milliseconds' and 'microseconds'.
        """
        s = self._date.isoformat() + " " + self._time.isoformat(timespec)

        return s

    __str__ = isoformat

    @classmethod
    def strptime(cls, date_string: str, format: str):
        bc = False
        if date_string.endswith(BCENDING):
            bc = True
            date_string = date_string.rstrip(BCENDING)
        if date_string.endswith(ADENDING):
            bc = False
            date_string = date_string.rstrip(ADENDING)
        date_string = date_string.strip()
        parsed_date = datetime.strptime(date_string, format)
        year = parsed_date.year if not bc else -parsed_date.year
        return cls(
            year,
            parsed_date.month,
            parsed_date.day,
            parsed_date.hour,
            parsed_date.minute,
            parsed_date.second,
            parsed_date.microsecond,
        )

    def strftime(self, format):
        date_string = datetime(
            abs(self.year),
            self.month,
            self.day,
            self.hour,
            self.minute,
            self.second,
            self.microsecond,
        ).strftime(format)
        if self.year < 0:
            date_string += BCENDING
        else:
            date_string += ADENDING

        return date_string



import unittest

from alldatetime.alldatetime import alldate, alldateperiod
from alldatetime.fuzzydatetime import fuzzydate, Precision, PrecisionUnit, fuzzydateperiod
from datetime import timedelta


class TestFuzzyDate(unittest.TestCase):
    def test_dateperiod(self):
        fuzzydates = [
            (fuzzydate(2023), alldateperiod(alldate(2023, 1, 1), alldate(2024, 1, 1))),
            (
                fuzzydate(2023, 1),
                alldateperiod(alldate(2023, 1, 1), alldate(2023, 1, 31)),
            ),
            (
                fuzzydate(2023, 1, 1),
                alldateperiod(alldate(2023, 1, 1), alldate(2023, 1, 2)),
            ),
            (
                fuzzydate(
                    2023, 1, 1, precision=Precision(num=2, unit=PrecisionUnit.Day)
                ),
                alldateperiod(alldate(2022, 12, 30), alldate(2023, 1, 3)),
            ),
        ]
        for fuzzy_date, date_period in fuzzydates:
            self.assertEqual(fuzzy_date.to_alldateperiod(), date_period)

    def test_overlap(self):
        fuzzydates = [
            (fuzzydate(202), fuzzydate(203), False),
            (fuzzydate(202), fuzzydate(202), True),
            (fuzzydate(202, 1), fuzzydate(203), False),
            (fuzzydate(202, 1), fuzzydate(202), True),
            (fuzzydate(202, 1, 1), fuzzydate(202), True),
            (fuzzydate(202, 12, 31), fuzzydate(202), True),
        ]
        for fuzzy_date1, fuzzy_date2, overlap in fuzzydates:
            self.assertEqual(fuzzy_date1.overlap_with(fuzzy_date2), overlap)
            self.assertEqual(fuzzy_date2.overlap_with(fuzzy_date1), overlap)

    def test_fuzzdateperiod(self):
        with self.assertRaises(ValueError):
            fuzzydateperiod(None, None)
        with self.assertRaises(ValueError):
            fuzzydateperiod(None, fuzzydate(202))
        with self.assertRaises(ValueError):
            fuzzydateperiod(fuzzydate(202), None)
        with self.assertRaises(ValueError):
            fuzzydateperiod(fuzzydate(202, 2), fuzzydate(202))
        period = fuzzydateperiod(fuzzydate(202, 1), fuzzydate(202, 5))
        self.assertTrue(period.cover(fuzzydate(202, 3)))
        self.assertTrue(period.cover(fuzzydate(202, 1)))
        self.assertFalse(period.cover(fuzzydate(202, 6)))
